from typing import List
import pandas as pd
from peperoncino import BaseProcessing


class StatsEncoding(BaseProcessing):
    """Encoding columns by statistical values of target column.

    Parameters
    ----------
    cols : List[str]
        A list of column names to be encoded.
    target : str
        The name of the target column.
    ops : List[str]
        A list of aggregation operation function names (e.g. ['mean', 'std']).
    ref : int
        A reference dataframe index to calculate the mapping from categories
        to encodings.
        Default values is 0(first dataframe).
    """

    def __init__(
        self, cols: List[str], target: str, ops: List[str], ref: int = 0,
    ):
        super().__init__(is_fixed_columns=False)
        self._cols = cols
        self._target = target
        self._ops = ops
        self._ref = ref

    def _process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        ref_df = dfs[self._ref]

        mapping = ref_df.groupby(self._cols)[self._target].agg(self._ops).reset_index()
        mapping = self._rename_op2col(mapping)

        dfs = [self._apply_mapping(df, mapping) for df in dfs]

        return dfs

    def _enc_names(self) -> List[str]:
        col_names = "&".join(self._cols)
        return [f"STATS_ENC_{col_names}_BY_{op}_{self._target}" for op in self._ops]

    def _rename_op2col(self, mapping: pd.DataFrame) -> pd.DataFrame:
        op2col = {op: name for op, name in zip(self._ops, self._enc_names())}
        return mapping.rename(columns=op2col)

    def _apply_mapping(self, df: pd.DataFrame, mapping: pd.DataFrame) -> pd.DataFrame:
        index = df.index

        # broadcast mapping to each row
        df = df.merge(
            mapping, on=self._cols, how="left", validate="many_to_one", sort=False,
        )
        df.index = index

        return df



from functools import partial
from itertools import permutations
from itertools import product
from itertools import combinations
from itertools import combinations_with_replacement
from typing import List
import pandas as pd

from peperoncino import SeparatedProcessing


class Combinations(SeparatedProcessing):
    """Calculate combination features from pairs of columns.
    The name of calculated features are discribed as Reverse Polish Notation.

    Parameters
    ----------
    cols : List[str]
        Columns to be used for combination calculation.
    ops : List[str]
        Calculation operations. e.g. ops=['*', '/']
    comb_type : str
        How to calculate combination features.
        Available options are:
        - combinaions
        - combinations_with_replacement
        - product
        - permutaions
        Please refer to `itertools` for more details.
    """

    def __init__(
        self,
        cols: List[str],
        ops: List[str],
        comb_type: str = "combinations_with_replacement",
    ):
        super().__init__(is_fixed_columns=False)
        self._cols = cols
        self._ops = ops

        if comb_type == "combinations":
            self._comb_fn = partial(combinations, r=2)
        elif comb_type == "combinations_with_replacement":
            self._comb_fn = partial(combinations_with_replacement, r=2)
        elif comb_type == "product":
            self._comb_fn = partial(product, repeat=2)
        elif comb_type == "permutations":
            self._comb_fn = partial(permutations, r=2)
        else:
            raise ValueError(
                "`comb_type` should be one of"
                " 'combinations', 'combinatinos_with_replacement',"
                " 'product' and 'permutations'"
            )

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        formulae = {}
        for a, b in self._comb_fn(self._cols):
            for op in self._ops:
                formulae[f"{op}_{a}_{b}"] = df.eval(f"{a} {op} {b}")
        return df.assign(**formulae)



from typing import Any
import pandas as pd
from peperoncino import SeparatedProcessing


class Assign(SeparatedProcessing):
    """Assign values to dataframes by given formula.
    scalar values are also applicable.

    e.g.
    ```
    import pp
    pp.Assign(
        foo = 'xxx * yyy',
        bar = 1,
    )
    ```

    Parameters
    ----------
    **formula : Any
        formula or values
    """

    def __init__(self, **formula: Any):
        super().__init__(is_fixed_columns=False)
        self._formula = formula

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        _assign = {}
        for k, f in self._formula.items():
            if isinstance(f, str):
                val = df.eval(f)
            else:
                val = f
            _assign[k] = val
        df = df.assign(**_assign)
        return df



from typing import Dict
import pandas as pd
from peperoncino import SeparatedProcessing


class RenameColumns(SeparatedProcessing):
    """Rename columns.

    Parameters
    ----------
    mapping: Dict[str, str]
        The name mapping.
    **kwargs:
        Keyward arguments are also concerned as the mapping.
    """

    def __init__(self, mapping: Dict[str, str] = {}, **kwargs: str):
        super().__init__(is_fixed_columns=False)
        mapping.update(kwargs)
        self._mapping = mapping

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.rename(self._mapping, axis=1)



from typing import List, Optional
import pandas as pd
from peperoncino import MergedProcessing


class AsCategory(MergedProcessing):
    """Change dtypes as category.

    Parameters
    ----------
    cols: List[str]
    fillna: Optional[str]
        If this is not None, missing values will be
        filled by this value and use it as category.
    """

    def __init__(self, cols: List[str], fillna: Optional[str] = None):
        super().__init__()
        self._cols = cols
        self._fillna = fillna

    def simul_process(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._fillna is not None:
            cols = df.get(self._cols)
            cols = cols.fillna(self._fillna)
            df = df.assign(**cols.to_dict("siries"))

        df = df.astype({c: "category" for c in self._cols})

        return df



from typing import Callable
import pandas as pd
from peperoncino import SeparatedProcessing


class ApplyColumn(SeparatedProcessing):
    """Apply function to a column.

    Parameters
    ----------
    col : str
    fn : Callable[[pd.Series], pd.Series]
        function to apply `col`.
    """

    def __init__(self, col: str, fn: Callable[[pd.Series], pd.Series]):
        super().__init__()
        self._col = col
        self._fn = fn

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(**{self._col: df[self._col].apply(self._fn)})
        return df



from typing import List
import pandas as pd
from peperoncino import SeparatedProcessing


class Select(SeparatedProcessing):
    def __init__(self, cols: List[str], lackable_cols: List[str] = []):
        super().__init__(is_fixed_columns=False)
        self._cols = cols
        self._lackable_cols = lackable_cols

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        cols = self._cols
        df_cols = df.columns.tolist()

        for c in self._cols:
            if c not in df.columns and c not in self._lackable_cols:
                raise ValueError(f"Column {c} must not be lacked.")

        cols = [c for c in cols if c in df_cols]
        return df[cols]



__version__ = "0.1.0"

from peperoncino.processing import BaseProcessing  # NOQA
from peperoncino.processing import SeparatedProcessing  # NOQA
from peperoncino.processing import MergedProcessing  # NOQA
from peperoncino.processing import ColumnsChangedError  # NOQA
from peperoncino.processing import RowsChangedError  # NOQA

from peperoncino.processings.pipeline import Pipeline  # NOQA
from peperoncino.processings.query import Query  # NOQA
from peperoncino.processings.apply_column import ApplyColumn  # NOQA
from peperoncino.processings.as_category import AsCategory  # NOQA
from peperoncino.processings.rename_columns import RenameColumns  # NOQA
from peperoncino.processings.assign import Assign  # NOQA
from peperoncino.processings.drop_columns import DropColumns  # NOQA
from peperoncino.processings.drop_duplicates import DropDuplicates  # NOQA
from peperoncino.processings.combinations import Combinations  # NOQA
from peperoncino.processings.select import Select  # NOQA
from peperoncino.processings.stats_encoding import StatsEncoding  # NOQA
from peperoncino.processings.target_encoding import TargetEncoding  # NOQA


def get_logger():  # type: ignore
    import logging
    import sys

    logger = logging.getLogger("peperoncino")
    logger.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter("[%(levelname)s]\tpeperoncino:\t%(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    return logger


logger = get_logger()



from __future__ import annotations
from abc import ABCMeta
from abc import abstractmethod
from typing import List, Set, Dict, Union, Any, Optional
import numpy as np
import pandas as pd

import peperoncino as pp


class ColumnsChangedError(Exception):
    pass


class RowsChangedError(Exception):
    pass


class BaseProcessing(metaclass=ABCMeta):
    """
    Abstruct class for data processing
    """

    def __init__(self, is_fixed_columns: bool = True, is_fixed_rows: bool = True):
        self._is_fixed_columns = is_fixed_columns
        self._is_fixed_rows = is_fixed_rows
        self._logs: List[Any] = []
        self._indices: Optional[List[int]] = None

    @property
    def is_fixed_columns(self) -> bool:
        return self._is_fixed_columns

    @property
    def is_fixed_rows(self) -> bool:
        return self._is_fixed_rows

    def _logging(self, msg: str, level: str = "info") -> None:
        if level not in ["fatal", "error", "warning", "info", "debug"]:
            raise ValueError(
                "`level` should be one of fatal, error, warning, info and debug"
            )
        self._logs.append((msg, level))

    def _flush_logs(self) -> None:
        for msg, level in self._logs:
            log_fn = getattr(pp.logger, level)
            log_fn(msg)
        self._logs.clear()

    def process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        """Processing dataframes

        Parameters
        ----------
        dfs : List[pd.DataFrame]

        Returns
        -------
        List[pd.DataFrame]
        """

        assert isinstance(dfs, list)
        for df in dfs:
            assert isinstance(df, pd.DataFrame)

        self._logging(f"Applying: {self.__class__.__name__}")

        # memory columns
        orig_cols = [df.columns for df in dfs]
        orig_rows = [df.index for df in dfs]

        dfs = self._process_with_limitation(dfs)

        cols = [df.columns for df in dfs]
        rows = [df.index for df in dfs]

        self._logging_summary(cols, rows, orig_cols, orig_rows)
        self._flush_logs()

        for i, (col, orig_col) in enumerate(zip(cols, orig_cols)):
            col, orig_col = set(col), set(orig_col)
            if self.is_fixed_columns and len(col | orig_col) != len(col & orig_col):
                raise ColumnsChangedError(
                    f"Number of columns are changed in df[{i}]."
                    f"Please refer to logs by setting proc.set_log_level(logging.DEBUG)"
                )

        for i, (row, orig_row) in enumerate(zip(rows, orig_rows)):
            if self.is_fixed_rows and len(row | orig_row) != len(row & orig_row):
                raise RowsChangedError(
                    f"Number of rows are changed in df[{i}]."
                    f"Please refer to logs by setting proc.set_log_level(logging.DEBUG)"
                )

        return dfs

    @abstractmethod
    def _process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        """Abstract method of processing

        Parameters
        ----------
        dfs : List[pd.DataFrame]

        Returns
        -------
        List[pd.DataFrame]
        """
        pass

    def _process_with_limitation(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        # limit dataframes by `only` function
        indices = self._indices
        if indices is None:
            indices = list(range(len(dfs)))

        _dfs = [df for i, df in enumerate(dfs) if i in indices]
        _dfs = self._process(_dfs)

        for i in range(len(dfs)):
            if i in indices:
                dfs[i] = _dfs.pop(0)

        return dfs

    def _logging_summary(
        self,
        cols: List[pd.Index],
        rows: List[pd.Index],
        orig_cols: List[pd.Index],
        orig_rows: List[pd.Index],
    ) -> None:
        for i, (col, row, orig_col, orig_row) in enumerate(
            zip(cols, rows, orig_cols, orig_rows)
        ):
            added_cols = set(col) - set(orig_col)
            dropped_cols = set(orig_col) - set(col)

            self._logging(f"df[{i}]")
            self._logging(f"#cols: {len(orig_col)} ---> {len(col)}")
            self._logging(f"+cols: {added_cols}", level="debug")
            self._logging(f"-cols: {dropped_cols}", level="debug")
            self._logging(f"#rows: {len(orig_row)} ---> {len(row)}")

    def only(self, indices: Union[int, List[int]]) -> BaseProcessing:
        """Limit the scope of processing

        Parameters
        ----------
        indices : Union[int, List[int]]

        Returns
        -------
        BaseProcessing
            self
        """

        if not isinstance(indices, int) or isinstance(indices, list):
            raise ValueError("indices must be int or list of ints")

        if isinstance(indices, int):
            indices = [indices]

        self._indices = indices
        return self


class SeparatedProcessing(BaseProcessing):
    """
    Abstract class to process each dataframe separatedly.
    """

    def _process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        return [self.sep_process(df) for df in dfs]

    @abstractmethod
    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Processing for single dataframe.

        Parameters
        ----------
        df : pd.DataFrame

        Returns
        -------
        pd.DataFrame
            A processed dataframe.
        """
        pass


class MergedProcessing(BaseProcessing):
    """
    Merge all dataframe and apply some processing simultaniously.
    """

    def _process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        orig_dtypes = self._gather_dtypes(dfs)
        dfs = [self._make_int_cols_nullable(df) for df in dfs]
        dtypes = self._gather_dtypes(dfs)
        df_indices = [df.index for df in dfs]

        xcols: List[List[str]] = []
        # set cols with valid dtype
        for i, df in enumerate(dfs):
            xcol = list(set(dtypes.keys()) - set(df.columns))
            df = df.assign(**{c: np.nan for c in xcol})
            df = df.astype(dtypes)
            dfs[i] = df
            xcols.append(xcol)

        # preserve dataframe ids and merge dataframes
        dfs = [df.assign(__DFID__=i) for i, df in enumerate(dfs)]
        merged_df = pd.concat(dfs, axis=0, sort=False).reset_index(drop=True)
        indices = [merged_df.query("__DFID__ == @i").index for i in range(len(dfs))]
        merged_df = merged_df.drop(columns="__DFID__")

        merged_df = self.simul_process(merged_df)

        for i, (index, df_index, xcol) in enumerate(zip(indices, df_indices, xcols)):
            df = merged_df.loc[index]
            df = df.drop(columns=xcol)
            df.index = df_index
            dfs[i] = df

        # Restore original dtype (only integers)
        restore_int = {}
        for c, dtype in dtypes.items():
            if not dtype.startswith("Int"):
                continue
            restore_int[c] = orig_dtypes[c]

        for i, df in enumerate(dfs):
            _dtypes = {k: v for k, v in restore_int.items() if k in df.columns}
            if len(_dtypes) > 0:
                dfs[i] = df.astype(_dtypes)

        return dfs

    def _gather_dtypes(self, dfs: List[pd.DataFrame]) -> Dict[str, str]:
        col_set: Set[str] = set()
        for df in dfs:
            col_set |= set(df.columns)

        dtypes: Dict[str, str] = {}
        for c in col_set:
            dtype_set: Set[str] = set()
            for df in dfs:
                if c not in df.columns:
                    continue
                dtype_set.add(str(df.dtypes[c]))

            if len(dtype_set) >= 2:
                raise ValueError(
                    f"Column {c} has different dtypes"
                    f"across given dataframes: {dtypes}"
                )

            dtypes[c] = dtype_set.pop()
        return dtypes

    def __process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        # make integer columns nullable
        dfs = [self._make_int_cols_nullable(df) for df in dfs]
        # preserve original index
        orig_indices = [df.index for df in dfs]
        # assign id and concat
        dfs = [df.assign(__DFID__=i) for i, df in enumerate(dfs)]
        merged_df = pd.concat(dfs, axis=0, sort=False).reset_index(drop=True)
        # preserve which indices are belonging to which df
        indices = [merged_df.query("__DFID__ == @i").index for i in range(len(dfs))]
        merged_df = merged_df.drop(columns="__DFID__")

        print(merged_df.dtypes)

        dropped_cols = []
        for df in dfs:
            # columns added by merging operation will be dropped
            dropped_cols.append(set(merged_df.columns) - set(df.columns))

        merged_df = self.simul_process(merged_df)

        for i, (df, index, orig_index) in enumerate(zip(dfs, indices, orig_indices)):
            # columns dropped in processing will be dropped too
            dcols = set(df.columns) - set(merged_df.columns)
            dcols |= dropped_cols[i]
            cols = [c for c in merged_df.columns if c not in dcols]

            df = merged_df.loc[index].get(cols)
            df.index = orig_index
            dfs[i] = df

        return dfs

    def _make_int_cols_nullable(self, df: pd.DataFrame) -> pd.DataFrame:
        dtypes = df.dtypes.astype(str)
        # select int
        dtypes = dtypes[dtypes.str.startswith("int")]
        # capitalized(e.g. Int64) int type accepts NaN
        dtypes = dtypes.str.capitalize()

        return df.astype(dtypes)

    @abstractmethod
    def simul_process(self, df: pd.DataFrame) -> pd.DataFrame:
        """Processing for the merged dataframe

        Parameters
        ----------
        df : pd.DataFrame

        Returns
        -------
        pd.DataFrame
        """
        pass



import pandas as pd
from peperoncino import SeparatedProcessing


class Query(SeparatedProcessing):
    """Query records in dataframe.

    Parameters
    ----------
    query : str
        query strings to be passed to pd.DataFrame.query.
    """

    def __init__(self, query: str):
        super().__init__(is_fixed_rows=False)
        self._query = query

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.query(self._query)



from typing import List
import pandas as pd
from peperoncino import BaseProcessing


class Pipeline(BaseProcessing):
    """Connect multiple processings

    Parameters
    ----------
    *procs : List[BaseProcessing]
    """

    def __init__(self, *procs: BaseProcessing):
        super().__init__(is_fixed_columns=False, is_fixed_rows=False)
        self._procs = procs

    def _process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        for p in self._procs:
            dfs = p.process(dfs)
        return dfs



from typing import List, Optional
import pandas as pd
from peperoncino import SeparatedProcessing


class DropDuplicates(SeparatedProcessing):
    """Drop duplicate rows.

    Parameters
    ----------
    cols : Optinal[List[str]]
        Only these columns are considered for uniqueness
        If this is None, all columns are considered.
        Default value is `None`.
    """

    def __init__(self, cols: Optional[List[str]] = None):
        super().__init__(is_fixed_rows=False)
        self._cols = cols

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        _df = df
        if self._cols is not None:
            _df = df.get(self._cols)
        index = _df.drop_duplicates().index
        return df.loc[index]



from typing import List
import pandas as pd
from peperoncino import BaseProcessing


def _broadcast_stats(stats: pd.DataFrame, df: pd.DataFrame) -> pd.DataFrame:
    """Broadcast aggregated statistical values to dataframe rows.

    Parameters
    ----------
    stats : pd.DataFrame
        dataframe generated by df.groupby(columns).agg(agg_list)
    df : pd.DataFrame
        original dataframe

    Returns
    -------
    pd.DataFrame
        Broadcasted stats
    """

    cols = list(stats.index.names)
    stats = (
        stats.reset_index()
        .merge(df[cols], on=cols, how="right", validate="one_to_many")
        .drop(columns=cols)
    )
    stats.index = df.index
    return stats


class TargetEncoding(BaseProcessing):
    """Target Mean Encoding with smoothing.
    https://dl.acm.org/citation.cfm?id=507538

    Parameters
    ----------
    cols : List[str]
        Columns to be encoded.
    target : str
        The target column.
    ref : int
        Reference dataframe index. Default is 0.
    prior_weight : float
        Prior weight for smoothing.
        If `prior_weight` > 0.0, the encoded values are calculated as
        `n_i / (n_i+prior_weight) * mean_i + prior_weight / (n+prior_weight) * prior`,
        where n_i is number of rows for category[i] and prior is mean of values.
        Default is 1.0.
    impute_prior : bool
        If True, categories which is not appeared in reference dataframe
        will be imputed by prior(the average of target).
        Default is True.
    """

    def __init__(
        self,
        cols: List[str],
        target: str,
        ref: int = 0,
        prior_weight: float = 1.0,
        impute_by_prior: bool = True,
    ):
        super().__init__(is_fixed_columns=False)
        self._cols = cols
        self._target = target
        self._ref = ref
        self._prior_weight = prior_weight
        self._impute_by_prior = impute_by_prior

    @property
    def enc_name(self) -> str:
        col_names = "&".join(self._cols)
        return f"TARGET_ENC_{col_names}_BY_{self._target}"

    def _process(self, dfs: List[pd.DataFrame]) -> List[pd.DataFrame]:
        ref_df = dfs[self._ref]

        # mapping
        mapping = ref_df.groupby(self._cols)[self._target].agg(["count", "mean"])

        # smoothing
        global_prior = ref_df[self._target].mean()

        if self._prior_weight > 0.0:
            prior = mapping.assign(count=len(ref_df), mean=global_prior)
            lamb = mapping["count"] / (mapping["count"] + self._prior_weight)
            mapping["mean"] = lamb * mapping["mean"] + (1 - lamb) * prior["mean"]

        mapping = mapping[["mean"]]
        dfs = [self._apply_mapping(df, mapping) for df in dfs]

        if self._impute_by_prior:
            dfs = [df.fillna({self.enc_name: global_prior}) for df in dfs]

        return dfs

    def _apply_mapping(self, df: pd.DataFrame, mapping: pd.DataFrame) -> pd.DataFrame:
        mapping = _broadcast_stats(mapping, df)

        # rename
        mapping = mapping.rename(columns={"mean": self.enc_name})

        df = df.merge(
            mapping,
            how="inner",
            left_index=True,
            right_index=True,
            sort=False,
            validate="one_to_one",
        )

        return df



from typing import List
import pandas as pd
from peperoncino import SeparatedProcessing


class DropColumns(SeparatedProcessing):
    """Drop columns.

    Parameters
    ----------
    cols : List[str]
    """

    def __init__(self, cols: List[str]):
        super().__init__(is_fixed_columns=False)
        self._cols = cols

    def sep_process(self, df: pd.DataFrame) -> pd.DataFrame:
        return df.drop(columns=self._cols)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['peperoncino', 'peperoncino.processings']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.17,<2.0', 'pandas>=0.25.3,<0.26.0']

setup_kwargs = {
    'name': 'peperoncino',
    'version': '0.0.1',
    'description': '',
    'long_description': None,
    'author': 'Junki Ishikawa',
    'author_email': '69guitar1015@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.７,<4.0',
}


setup(**setup_kwargs)
